/*
 * Copyright (c) 2019-2021, AT&T Intellectual Property.  All rights reserved.
 *
 * SPDX-License-Identifier: LGPL-2.1-only
 */

/**
 * @file cgn_cmd_cfg.c - CGNAT config
 *
 * -----------------------------------------------
 * Policy config
 * -----------------------------------------------
 *
 * cgn-cfg policy add <policy-name>
 *   priority=<priority>
 *   src-addr=<prefix/length>
 *   pool=<pool-name>
 *   log-group=<group-name>
 *   log-all={yes | no}
 *
 * cgn-cfg policy delete <policy-name>
 * cgn-cfg policy attach <policy-name> <interface-name>
 * cgn-cfg policy detach <policy-name> <interface-name>
 *
 * -----------------------------------------------
 * Event config
 * -----------------------------------------------
 *
 * cgn-cfg events rte_log|protobuf <type> enable|disable
 * cgn-cfg events protobuf <type> hwm <integer>
 * cgn-cfg events core [<core-num>]
 *
 * -----------------------------------------------
 * Other config
 * -----------------------------------------------
 *
 * cgn-cfg hairpinning {on | off}
 * cgn-cfg snat-alg-bypass {on | off}
 */

#include <errno.h>
#include <netinet/in.h>
#include <linux/if.h>

#include "commands.h"
#include "compiler.h"
#include "config_internal.h"
#include "if_var.h"
#include "util.h"
#include "vplane_log.h"

#include "npf/npf_addr.h"

#include "npf/cgnat/cgn.h"
#include "npf/cgnat/cgn_if.h"
#include "npf/cgnat/cgn_limits.h"
#include "npf/cgnat/cgn_policy.h"
#include "npf/cgnat/cgn_sess_state.h"
#include "npf/cgnat/cgn_session.h"
#include "npf/cgnat/cgn_cmd_cfg.h"
#include "npf/cgnat/cgn_log.h"
#include "npf/cgnat/cgn_log_protobuf_zmq.h"
#include "npf/cgnat/alg/alg_public.h"

#include "npf/apm/apm.h"
#include "npf/nat/nat_pool.h"
#include "npf/cgnat/cgn_session.h"
#include "npf/cgnat/cgn_source.h"


/*
 * Extract an integer from a string
 */
int cgn_arg_to_int(const char *arg)
{
	char *p;
	unsigned long val = strtoul(arg, &p, 10);

	if (p == arg || val > INT_MAX)
		return -1;

	return (uint32_t) val;
}

/*
 * Iterate through argv/argc looking for "intf=dp0p1", extract the interface
 * name, and lookup the ifp pointer.  Does not change argv.
 */
static char *
cgn_cfg_ifname_from_arg(char *if_name, int sz, int argc, char **argv)
{
	char *c, *item, *value;
	int i;

	if (sz < IFNAMSIZ)
		return NULL;

	for (i = 0; i < argc; i++) {
		if (!strstr(argv[i], "intf="))
			continue;

		c = strchr(argv[i], '=');
		if (!c)
			return NULL;

		/* Duplicate the string so we can write to it */
		item = strdup(argv[i]);
		if (!item)
			return NULL;

		c = strchr(item, '=');
		if (!c) {
			free(item);
			return NULL;
		}

		*c = '\0';
		value = c + 1;

		strncpy(if_name, value, sz-1);
		if_name[sz - 1] = '\0';

		free(item);
		return if_name;
	}
	return NULL;
}

/*
 * Attach policy to interface
 *
 * cgn-cfg policy attach intf=dp0p1 name=POLICY1
 */
static int cgn_policy_cfg_attach(FILE *f, int argc, char **argv)
{
	const char *name = NULL;
	struct ifnet *ifp;
	struct cgn_policy *cp;
	char ifname[IFNAMSIZ+1];
	char *c, *item, *value;
	int i;

	if (argc < 5)
		goto usage;

	/* Extract interface name from "intf=dp0p1" arg list */
	if (!cgn_cfg_ifname_from_arg(ifname, sizeof(ifname), argc, argv))
		goto usage;

	/* Does interface exist? */
	ifp = dp_ifnet_byifname(ifname);

	if (!ifp) {
		RTE_LOG(ERR, CGNAT, "Interfaace %s not found\n", ifname);
		goto err_out;
	}

	/*
	 * Parse item/value pairs.  We ignore any we do not understand.
	 */
	for (i = 0; i < argc; i++) {
		c = strchr(argv[i], '=');
		if (!c)
			continue;

		item = argv[i];
		*c = '\0';
		value = c + 1;

		if (!strcmp(item, "intf"))
			continue;

		if (!strcmp(item, "name"))
			name = value;
	}

	if (!name)
		goto usage;

	cp = cgn_policy_lookup(name);
	if (!cp)
		return -EEXIST;

	/* Add policy to cgn interface list, and take reference on policy */
	cgn_if_add_policy(ifp, cp);

	return 0;

usage:
	if (f)
		fprintf(f, "%s: policy attach name=<policy-name> "
			"intf=<intf-name>",
			__func__);
err_out:
	return -1;
}

/*
 * Detach policy from interface
 *
 * cgn policy detach name=POLICY1 intf=dpT21
 */
static int cgn_policy_cfg_detach(FILE *f, int argc, char **argv)
{
	struct cgn_policy *cp;
	const char *name = NULL;
	char *ifname = NULL;
	struct ifnet *ifp;
	char *c, *item, *value;
	int i;

	if (argc < 5)
		goto usage;

	/*
	 * Parse item/value pairs.  We ignore any we do not understand.
	 */
	for (i = 0; i < argc; i++) {
		c = strchr(argv[i], '=');
		if (!c)
			continue;

		item = argv[i];
		*c = '\0';
		value = c + 1;

		if (!strcmp(item, "intf"))
			ifname = value;

		else if (!strcmp(item, "name"))
			name = value;
	}

	if (!name || !ifname)
		goto usage;

	/* Does interface exist? */
	ifp = dp_ifnet_byifname(ifname);
	if (!ifp)
		return -EEXIST;

	/*
	 * Policy may have been removed from the hash table before now, so
	 * search list
	 */
	cp = cgn_if_find_policy_by_name(ifp, name);
	if (!cp)
		return 0;

	/*
	 * Delete policy from interface list and release reference on policy
	 */
	cgn_if_del_policy(ifp, cp);

	return 0;

usage:
	if (f)
		fprintf(f, "%s: policy detach <policy-name> <intf-name>",
			__func__);

	return -1;
}

/*
 * cgn-cfg policy ...
 */
static int cgn_policy_cfg(FILE *f, int argc, char **argv)
{
	int rc = 0;

	if (argc < 3)
		goto usage;

	/* Policy */
	if (strcmp(argv[2], "add") == 0)
		rc = cgn_policy_cfg_add(f, argc, argv);

	else if (strcmp(argv[2], "delete") == 0)
		rc = cgn_policy_cfg_delete(f, argc, argv);

	else if (strcmp(argv[2], "attach") == 0)
		rc = cgn_policy_cfg_attach(f, argc, argv);

	else if (strcmp(argv[2], "detach") == 0)
		rc = cgn_policy_cfg_detach(f, argc, argv);
	else
		goto usage;

	return rc;
usage:
	if (f)
		fprintf(f, "%s: cgn-cfg policy {add|delete} ... ",
			__func__);

	return -1;
}

/*
 * cgn-cfg events rte_log <type> enable|disable
 *
 * <type> is one of session, port-block-allocation, subscriber,
 * or resource-constraint
 */
static int cgn_events_cfg_rte_log(FILE *f, int argc, char **argv)
{
	const char *ltype_str;
	enum cgn_log_type ltype;
	int rc;

	if (argc < 5) {
		if (f)
			fprintf(f, "%s: need at least 5 fields", __func__);
		return -1;
	}

	ltype_str = argv[3];

	rc = cgn_get_log_type(ltype_str, &ltype);
	if (rc < 0) {
		if (f)
			fprintf(f, "%s: unknown event type %s", __func__,
				ltype_str);
		return -1;
	}

	if (strcmp(argv[4], "enable") == 0) {
		rc = cgn_log_enable_handler(ltype, "rte_log");
		if (rc < 0 && rc != -EEXIST) {
			if (f)
				fprintf(f, "%s: cgn_log_enable_handler failed "
					"for type %s", __func__, ltype_str);
			return -1;
		}
	} else if (strcmp(argv[4], "disable") == 0) {
		rc = cgn_log_disable_handler(ltype, "rte_log");
		if (rc < 0 && rc != -ENOENT) {
			if (f)
				fprintf(f, "%s: cgn_log_disable_handler failed "
					"for type %s", __func__, ltype_str);
			return -1;
		}
	} else {
		if (f)
			fprintf(f, "%s: unexpected value %s for type %s",
				__func__, argv[4], ltype_str);
		return -1;
	}

	return 0;
}

/*
 * cgn-cfg events protobuf <type> enable|disable|hwm
 *
 * <type> is one of session, port-block-allocation, subscriber,
 * or resource-constraint
 */
static int cgn_events_cfg_protobuf(FILE *f, int argc, char **argv)
{
	const char *ltype_str;
	enum cgn_log_type ltype;
	int rc;

	if (argc < 5) {
		if (f)
			fprintf(f, "%s: need at least 5 fields", __func__);
		return -1;
	}

	ltype_str = argv[3];

	rc = cgn_get_log_type(ltype_str, &ltype);
	if (rc < 0) {
		if (f)
			fprintf(f, "%s: unknown event type %s", __func__,
				ltype_str);
		return -1;
	}

	if (strcmp(argv[4], "enable") == 0) {
		rc = cgn_log_enable_handler(ltype, "protobuf");
		if (rc < 0 && rc != -EEXIST) {
			if (f)
				fprintf(f, "%s: cgn_log_enable_handler failed "
					"for type %s", __func__, ltype_str);
			return -1;
		}
	} else if (strcmp(argv[4], "disable") == 0) {
		rc = cgn_log_disable_handler(ltype, "protobuf");
		if (rc < 0 && rc != -ENOENT) {
			if (f)
				fprintf(f, "%s: cgn_log_disable_handler failed "
					"for type %s", __func__, ltype_str);
			return -1;
		}
	} else if (strcmp(argv[4], "hwm") == 0) {
		int32_t hwm;

		if (argc >= 6)
			hwm = atoi(argv[5]);
		else
			hwm = 0;	/* default of unlimited */

		rc = cl_zmq_set_hwm(ltype, hwm);
		if (rc < 0) {
			if (f)
				fprintf(f, "%s: cl_zmq_set_hwm failed "
					"for type %s, hwm \"%s\"",
					__func__, ltype_str,
					argc >= 6 ? argv[5] : "default");
			return -1;
		}
	} else {
		if (f)
			fprintf(f, "%s: unexpected value %s for type %s",
				__func__, argv[4], ltype_str);
		return -1;
	}

	return 0;
}

/*
 * cgn-cfg events core [<core-num>]
 *
 * <core-num> is the number of the core requested to handle session log/export
 * events as a thread on the dedicated core
 */
static int cgn_events_cfg_core(FILE *f, int argc, char **argv)
{
	int rc;

	if (argc >= 4) {
		int core_num = cgn_arg_to_int(argv[3]);

		if (core_num < 0) {
			fprintf(f, "%s: core number cannot be negative",
				__func__);
			return -1;
		}

		rc = cgn_set_helper_thread((unsigned int) core_num);
		if (rc < 0 && f)
			fprintf(f, "%s: cgn_set_helper_thread failed for "
				"core %s", __func__, argv[3]);
		/*
		 * NB: do not fail, to prevent dataplane from continually
		 * restarting if the configuration is replayed, as
		 * unsupported CPU numbers could be configured.
		 */
	} else {
		rc = cgn_disable_helper_thread();
		if (rc < 0) {
			if (f)
				fprintf(f, "%s: cgn_disable_helper_thread "
					"failed", __func__);
			return -1;
		}
	}

	return 0;
}

/*
 * cgn-cfg events rte_log|protobuf
 */
static int cgn_events_cfg(FILE *f, int argc, char **argv)
{
	int rc = 0;

	if (argc < 3)
		goto usage;

	if (strcmp(argv[2], "rte_log") == 0)
		rc = cgn_events_cfg_rte_log(f, argc, argv);

	else if (strcmp(argv[2], "protobuf") == 0)
		rc = cgn_events_cfg_protobuf(f, argc, argv);

	else if (strcmp(argv[2], "core") == 0)
		rc = cgn_events_cfg_core(f, argc, argv);

	else
		goto usage;

	return rc;

usage:
	if (f)
		fprintf(f, "%s: cgn-cfg events {rte_log|protobuf|core} ... ",
			__func__);

	return -1;
}

/*
 * cgn-cfg hairpinning [on|off]
 */
static int cgn_hairpinning_cfg(FILE *f, int argc, char **argv)
{
	if (argc < 3)
		goto usage;

	/* Policy */
	if (strcmp(argv[2], "on") == 0)
		cgn_hairpinning_gbl = true;
	else
		cgn_hairpinning_gbl = false;

	return 0;
usage:
	if (f)
		fprintf(f, "%s: cgn-cfg hairpinning {on|off}",
			__func__);

	return -1;
}

/*
 * cgn-cfg snat-alg-bypass [on|off]
 */
static int cgn_snat_alg_bypass_cfg(FILE *f, int argc, char **argv)
{
	if (argc < 3)
		goto usage;

	/* Policy */
	if (strcmp(argv[2], "on") == 0)
		cgn_snat_alg_bypass_gbl = true;
	else
		cgn_snat_alg_bypass_gbl = false;

	return 0;
usage:
	if (f)
		fprintf(f, "%s: cgn-cfg snat-alg-bypass {on|off}",
			__func__);

	return -1;
}

/*
 * cgn-cfg alg {sip | ftp | pptp} {on | off}
 */
static int cgn_alg_cfg(FILE *f, int argc, char **argv)
{
	int rc;

	if (argc < 4)
		goto usage;

	if (strcmp(argv[3], "on") == 0)
		rc = cgn_alg_enable(argv[2]);
	else
		rc = cgn_alg_disable(argv[2]);

	if (rc == 0)
		return 0;
usage:
	if (f)
		fprintf(f, "%s: cgn-cfg alg {sip|ftp|pptp} {on|off}",
			__func__);

	return -1;
}

/*
 * cgn-cfg max-sessions <num>
 */
static int cgn_max_sessions_cfg(FILE *f, int argc, char **argv)
{
	int tmp;

	if (argc < 3)
		goto usage;

	tmp = cgn_arg_to_int(argv[2]);
	if (tmp < 0 || tmp > CGN_SESSIONS_MAX)
		return -1;

	if (tmp == 0)
		tmp = CGN_SESSIONS_MAX;
	cgn_session_set_max(tmp);

	return 0;
usage:
	if (f)
		fprintf(f, "%s: cgn-cfg max-sessions <num>",
			__func__);

	return -1;
}

static int
cgn_max_apms_cfg(FILE *f __unused, int argc __unused, char **argv __unused)
{
	/* Deprecated */
	return 0;
}

/*
 * cgn-cfg max-subscribers <num>
 */
static int cgn_max_subscribers_cfg(FILE *f, int argc, char **argv)
{
	int tmp;

	if (argc < 3)
		goto usage;

	tmp = cgn_arg_to_int(argv[2]);
	if (tmp < 0 || tmp > CGN_SRC_TABLE_MAX)
		return -1;

	if (tmp == 0)
		tmp = CGN_SRC_TABLE_MAX;

	cgn_source_set_max(tmp);

	return 0;
usage:
	if (f)
		fprintf(f, "%s: cgn-cfg max-subscribers <num>",
			__func__);

	return -1;
}

/*
 * cgn-cfg max-dest-per-session <num>
 *
 * cs_sess2_used is a an atomic int16, so the value cgn_dest_sessions_max must
 * never be greater than USHRT_MAX - 1 to avoid wrap.
 */
static int cgn_max_dest_sessions_cfg(FILE *f, int argc, char **argv)
{
	uint16_t tmp;

	if (argc < 3)
		goto usage;

	tmp = (uint16_t)cgn_arg_to_int(argv[2]);
	if (tmp > CGN_DEST_SESSIONS_MAX)
		return -1;

	if (tmp == 0)
		tmp = CGN_DEST_SESSIONS_INIT;

	/*
	 * cgn_dest_sessions_max is used to limit the number of entries to the
	 * dest session hash table.
	 */
	cgn_dest_sessions_max = tmp;

	/*
	 * cgn_dest_ht_max is used to initialise the dest session hash table.
	 * This must be a power of two, so we round up tmp accordingly.
	 */
	for (uint16_t po2 = CGN_DEST_SESSIONS_MAX >> 1; po2 > 0; po2 >>= 1) {
		if (tmp > po2) {
			tmp = po2 << 1;
			break;
		}
	}
	cgn_dest_ht_max = tmp;

	return 0;
usage:
	if (f)
		fprintf(f, "%s: cgn-cfg max-dest-per-session <num>",
			__func__);

	return -1;
}

/*
 * Remaining command is one off:
 *
 *   tcp-estab <timeout>
 *   tcp-estab port <port> timeout <timeout>
 *
 * Returns the number of arguments consumed, or 0 if there is an error.
 */
static int cgn_sess_timeout_tcp_estbd(int argc, char **argv)
{
	int port, timeout;

	if (!strcmp(argv[1], "port")) {
		if (argc < 5)
			return 0;

		port = cgn_arg_to_int(argv[2]);
		if (port < 0 || port > USHRT_MAX)
			return 0;

		if (strcmp(argv[3], "timeout") != 0)
			return 0;

		timeout = cgn_arg_to_int(argv[4]);
		if (timeout < 0)
			return 0;

		cgn_cgn_port_tcp_etime_set(port, timeout);

		/* Five args consumed */
		return 5;
	}

	timeout = cgn_arg_to_int(argv[1]);
	if (timeout < 0)
		return 0;

	cgn_sess_tcp_etime[CGN_ETIME_TCP_ESTBD] = timeout;

	/* Two args consumed */
	return 2;
}

/*
 * Remaining command is one off:
 *
 *   udp-estab <timeout>
 *   udp-estab port <port> timeout <timeout>
 *
 * Returns the number of arguments consumed, or 0 if there is an error.
 */
static int cgn_sess_timeout_udp_estbd(int argc, char **argv)
{
	int port, timeout;

	if (!strcmp(argv[1], "port")) {
		if (argc < 5)
			return 0;

		port = cgn_arg_to_int(argv[2]);
		if (port < 0 || port > USHRT_MAX)
			return 0;

		if (strcmp(argv[3], "timeout") != 0)
			return 0;

		timeout = cgn_arg_to_int(argv[4]);
		if (timeout < 0)
			return 0;

		cgn_cgn_port_udp_etime_set(port, timeout);

		/* Five args consumed */
		return 5;
	}

	timeout = cgn_arg_to_int(argv[1]);
	if (timeout < 0)
		return 0;

	cgn_sess_udp_etime[CGN_ETIME_ESTBD] = timeout;

	/* Two args consumed */
	return 2;
}

/*
 * Session timeouts
 */
static int cgn_session_timeouts_cfg(FILE *f, int argc, char **argv)
{
	/* Move past "cgn-cfg session-timeouts" */
	argc -= 2;
	argv += 2;

	if (argc < 2)
		goto usage;

	/*
	 * Parse item/value pairs.  We ignore any we do not understand.
	 */
	while (argc > 0) {
		char *item;
		int tmp;

		item = argv[0];

		if (!strcmp(item, "other-opening") && argc >= 2) {
			tmp = cgn_arg_to_int(argv[1]);
			if (tmp < 0)
				goto invalid_value;

			cgn_sess_other_etime[CGN_ETIME_OPENING] = tmp;

			argc -= 2;
			argv += 2;

		} else if (!strcmp(item, "other-estab") && argc >= 2) {
			tmp = cgn_arg_to_int(argv[1]);
			if (tmp < 0)
				goto invalid_value;

			cgn_sess_other_etime[CGN_ETIME_ESTBD] = tmp;

			argc -= 2;
			argv += 2;

		} else if (!strcmp(item, "udp-opening") && argc >= 2) {
			tmp = cgn_arg_to_int(argv[1]);
			if (tmp < 0)
				goto invalid_value;

			cgn_sess_udp_etime[CGN_ETIME_OPENING] = tmp;

			argc -= 2;
			argv += 2;

		} else if (!strcmp(item, "udp-estab") && argc >= 2) {
			tmp = cgn_sess_timeout_udp_estbd(argc, argv);

			if (tmp == 0)
				goto invalid_value;

			argc -= tmp;
			argv += tmp;

		} else if (!strcmp(item, "tcp-opening") && argc >= 2) {
			tmp = cgn_arg_to_int(argv[1]);
			if (tmp < 0)
				goto invalid_value;

			cgn_sess_tcp_etime[CGN_ETIME_TCP_OPENING] = tmp;

			argc -= 2;
			argv += 2;

		} else if (!strcmp(item, "tcp-estab") && argc >= 2) {
			tmp = cgn_sess_timeout_tcp_estbd(argc, argv);

			if (tmp == 0)
				goto invalid_value;

			argc -= tmp;
			argv += tmp;

		} else if (!strcmp(item, "tcp-closing") && argc >= 2) {
			tmp = cgn_arg_to_int(argv[1]);
			if (tmp < 0)
				goto invalid_value;

			cgn_sess_tcp_etime[CGN_ETIME_TCP_CLOSING] = tmp;

			argc -= 2;
			argv += 2;

		} else
			goto usage;
	}

	return 0;

usage:
	if (f)
		fprintf(f, "%s: cgn-cfg "
			"session-timeouts <item> <value> ...",
			__func__);

	return -1;

invalid_value:
	if (f)
		fprintf(f, "%s: cgn-cfg "
			"session-timeouts invalid value %s",
			__func__, argv[1]);

	return -1;
}

/* Warning threshold setter functions */
typedef void (*threshold_func_t)(int32_t threshold, uint32_t interval);

struct threshold_fn_t {
	const char *name;
	threshold_func_t fn;
};

static const struct threshold_fn_t threshold_fns[] = {
	{ "mapping-table", apm_table_threshold_set },
	{ "session-table", session_table_threshold_set },
	{ "subscriber-table", subscriber_table_threshold_set },
	{ "public-addresses", np_threshold_set_all },
};

/* Configure CGN threshold warning levels */
static int cgn_threshold_cfg(FILE *f, int argc, char **argv)
{
	bool add;

	if (argc < 3)
		goto usage;

	if (strcmp(argv[2], "add") == 0)
		add = true;

	else if (strcmp(argv[2], "del") == 0)
		add = false;

	else
		goto usage;

	/*
	 * Expecting "cgn-config warning del NNNN"
	 *        or "cgn-config warning add NNNN threshold TTT"
	 *        or "cgn-config warning add NNNN threshold TTT interval III"
	 */
	if (argc < 4)
		return -EINVAL;

	char *name = argv[3];
	int32_t interval = 0;
	int32_t threshold = 0;

	if (add) {
		if (argc < 6)
			return -EINVAL;

		assert(!strcmp("threshold", argv[4]));
		threshold = atoi(argv[5]);

		if (argc >= 8) {
			assert(!strcmp("interval", argv[6]));
			interval = atoi(argv[7]);
		}
	}

	for (uint32_t i = 0; i < ARRAY_SIZE(threshold_fns); i++) {
		if (!strcmp(name, threshold_fns[i].name)) {
			threshold_fns[i].fn(threshold, interval);
			return 0;
		}
	}

	/* No function found */
	return -1;

usage:
	if (f)
		fprintf(f, "%s: cgn-cfg warning {add|delete} ... ",
			__func__);

	return -1;
}

/*
 * cgn-cfg [<type>] ...
 * cgn-ut  ...
 */
int cmd_cgn(FILE *f, int argc, char **argv)
{
	int rc = 0;

	if (argc < 2)
		goto usage;

	if (strcmp(argv[1], "policy") == 0)
		rc = cgn_policy_cfg(f, argc, argv);

	else if (strcmp(argv[1], "events") == 0)
		rc = cgn_events_cfg(f, argc, argv);

	else if (strcmp(argv[1], "hairpinning") == 0)
		rc = cgn_hairpinning_cfg(f, argc, argv);

	else if (strcmp(argv[1], "snat-alg-bypass") == 0)
		rc = cgn_snat_alg_bypass_cfg(f, argc, argv);

	else if (strcmp(argv[1], "alg") == 0)
		rc = cgn_alg_cfg(f, argc, argv);

	else if (strcmp(argv[1], "max-sessions") == 0)
		rc = cgn_max_sessions_cfg(f, argc, argv);

	else if (strcmp(argv[1], "max-apms") == 0)
		rc = cgn_max_apms_cfg(f, argc, argv);

	else if (strcmp(argv[1], "max-subscribers") == 0)
		rc = cgn_max_subscribers_cfg(f, argc, argv);

	else if (strcmp(argv[1], "max-dest-per-session") == 0)
		rc = cgn_max_dest_sessions_cfg(f, argc, argv);

	else if (strcmp(argv[1], "session-timeouts") == 0)
		rc = cgn_session_timeouts_cfg(f, argc, argv);

	else if (strcmp(argv[1], "warning") == 0)
		rc = cgn_threshold_cfg(f, argc, argv);

	else
		goto usage;

	return rc;

usage:
	if (f) {
		if (argc < 2)
			fprintf(f, "%s: cgn-cfg with no type parameter",
				__func__);
		else
			fprintf(f, "%s: cgn-cfg with unknown type: %s",
				__func__, argv[1]);
	}

	return -1;
}

int cmd_cgn_ut(FILE *f, int argc, char **argv)
{
	return cmd_cgn(f, argc, argv);
}
